use std::collections::{HashMap, HashSet};

use crate::grammar::Grammar;
use crate::utils::MapOfSet;

#[derive(Debug, Default)]
pub struct ContextSets {}

#[derive(Debug, Error)]
pub enum Error {}

impl Grammar {
    pub fn build_context_sets(&mut self) {
        // first set construction: dragon book page 189

        let mut first_sets = MapOfSet::<String, String>::default();
        let grammar_symbols = self.symbol_names();

        loop {
            let mut to_add = HashSet::<(String, String)>::new();

            for name in grammar_symbols.iter() {
                let name = *name;
                if self.is_terminal(name) {
                    if !first_sets.key_contains(name, name) {
                        to_add.insert((name.to_owned(), name.to_owned()));
                    }
                } else if self.is_nonterminal(name) {
                    let x = name;
                    let productions = self.productions.get(name).unwrap();
                    for production in productions.iter() {
                        // has ε shown up in every FIRST set so far?
                        let mut is_epsilon_in_all_sets = false;
                        let mut first = true;

                        // if X -> ε is a production
                        if production.len() == 1 && production[0] == "ε" {
                            if !first_sets.key_contains(x, "ε") {
                                to_add.insert((
                                    x.to_owned(),
                                    "ε".to_owned(),
                                ));
                            }
                        }

                        if first && !is_epsilon_in_all_sets {
                            first = false;
                            is_epsilon_in_all_sets = true;
                        }

                        // if X is a terminal and X -> Y1 Y2 Y3 .. Yk
                        for (_i, y_i) in production.iter().enumerate() {
                            if let Some(first_y) = first_sets.get(y_i) {
                                for a in first_y {
                                    if is_epsilon_in_all_sets
                                        && !first_sets.key_contains(x, a)
                                    {
                                        to_add.insert((
                                            x.to_owned(),
                                            a.to_owned(),
                                        ));
                                    }
                                }
                            }

                            if let Some(first_y) = first_sets.get(y_i) {
                                if !first_y.contains("ε") {
                                    is_epsilon_in_all_sets = false;
                                }
                            } else {
                                is_epsilon_in_all_sets = false;
                            }
                        }

                        if is_epsilon_in_all_sets {
                            // ε is in all first sets for X's production
                            if !first_sets.key_contains(x, "ε") {
                                to_add.insert((
                                    x.to_owned(),
                                    "ε".to_owned(),
                                ));
                            }
                        }
                    }
                }
            }

            if to_add.is_empty() {
                break;
            } else {
                for (symbol_name, first_item) in to_add.drain() {
                    first_sets.insert_item(symbol_name, first_item);
                }
            }
        }

        self.first_sets = first_sets;
    }
}

#[cfg(test)]
mod tests {
    use anyhow::Result;

    use super::ContextSets;

    macro_rules! assert_first_sets {
        (
            $set:expr,
            $(
                $lhs:ident
                ->
                $($arg:ident)*
             ),* $(,)?
        ) => {
            $(
                let first_set = $set.get(stringify!($lhs)).expect("should exist");
                $(
                    assert!(first_set.contains(stringify!($arg)));
                 )*
             )*
        }
    }

    #[test]
    fn example1() -> Result<()> {
        let mut grammar = make_grammar! {
            E -> T E_,
            E_ -> P T E_ | ε,
            T -> F T_,
            T_ -> A F T_ | ε,
            F -> LP E RP | num
            %
            P -> r"\+",
            A -> r"\*",
            LP -> r"\(",
            RP -> r"\)",
            num -> r"[0-9]+",
        };
        grammar.build_context_sets();

        assert_first_sets! {
            grammar.first_sets,
            E -> LP num,
            T -> LP num,
            F -> LP num,
            E_ -> P ε,
            T_ -> A ε,
        };

        Ok(())
    }

    #[test]
    fn example2() -> Result<()> {
        let mut grammar = make_grammar! {
            S -> A C B | C b b | B a,
            A -> d a | B C,
            B -> g | ε,
            C -> h | ε,
            %
            a -> "a",
            b -> "b",
            d -> "d",
            g -> "g",
            h -> "h",
        };
        grammar.build_context_sets();

        assert_first_sets! {
            grammar.first_sets,
            S -> a b d g h ε,
            A -> d g h ε,
            B -> g ε,
            C -> h ε,
        };

        Ok(())
    }
}
