mod context_sets;

use std::collections::{HashMap, HashSet};

use regex::Regex;

use crate::canonical_collection::CanonicalCollection;
use crate::lr0item::Lr0Item;
use crate::utils::MapOfSet;

#[derive(Debug)]
pub struct Grammar {
    /// The start symbols
    pub start: HashSet<String>,

    /// Map of terminals from their names to the regex used to identify
    /// them
    pub terminals: HashMap<String, Regex>,

    pub productions: HashMap<String, Vec<Vec<String>>>,

    pub first_sets: MapOfSet<String, String>,
}

impl Grammar {
    pub fn symbol_names(&self) -> Vec<&str> {
        self.terminals
            .keys()
            .map(|s| s.as_str())
            .chain(self.productions.keys().map(|s| s.as_str()))
            .collect()
    }

    pub fn is_terminal(&self, name: impl AsRef<str>) -> bool {
        self.terminals.contains_key(name.as_ref())
    }

    pub fn is_nonterminal(&self, name: impl AsRef<str>) -> bool {
        self.productions.contains_key(name.as_ref())
    }

    pub fn build(&mut self) -> CanonicalCollection {
        self.build_context_sets();

        let sets = HashSet::new();

        CanonicalCollection {
            grammar: self,
            sets,
        }
    }

    pub fn closure(&self, I: HashSet<Lr0Item>) -> HashSet<Lr0Item> {
        // closure described in dragon book page 223
        let mut J = I.clone();
        let mut to_add = HashSet::new();
        loop {
            for item in J.iter() {
                let A = &item.lhs;
                let B = item.next_symbol();
                let productions = match self.productions.get(B) {
                    Some(v) => v,
                    None => continue,
                };
                for production in productions {
                    let item = Lr0Item {
                        lhs: B.to_owned(),
                        rhs: production.clone(),
                        dot: 0,
                    };

                    if !J.contains(&item) {
                        to_add.insert(item);
                    }
                }
            }

            if to_add.is_empty() {
                break;
            } else {
                J.extend(to_add.drain());
            }
        }
        J
    }

    pub fn goto(&self, I: HashSet<Lr0Item>, X: &str) -> HashSet<Lr0Item> {
        let mut result = HashSet::new();
        let mut to_add = HashSet::new();
        loop {
            for item in I.iter() {
                let A = &item.lhs;
                let B = item.next_symbol();
                let productions = match self.productions.get(B) {
                    Some(v) => v,
                    None => continue,
                };
                for production in productions {}
            }

            if to_add.is_empty() {
                break;
            } else {
                result.extend(to_add.drain());
            }
        }
        return self.closure(result);
    }
}

#[cfg(test)]
mod tests {
    use crate::lr0item::Lr0Item;

    #[test]
    fn closure1() {
        let grammar = make_grammar! {
            E_ -> E,
            E -> E p T | T,
            T -> T a F | F,
            F -> lp E rp | num,
            %
            p -> r"\+",
            a -> r"\*",
            lp -> r"\(",
            rp -> r"\)",
            num -> r"[0-9]+",
        };

        let item = Lr0Item {
            lhs: string!(E_),
            rhs: vec![string!(E)],
            dot: 0,
        };

        let j = grammar.closure(hashset! { item });
        // TODO: verify this later
        // E_ -> . E
        // E -> . E p T
        // E -> . T
        // T -> . T a F
        // T -> . F
        // F -> . lp E rp
        // F -> . num
    }
}
